/*
**  Facility:
**
**      safe_shell_exec.c
**
**  Abstract:
**
**      Provides an interface similar to the PHP native shell_exec()
**      without the need for executing a shell process.
**
**  Authors:
**
**      Dave McMurtrie <dave64@andrew.cmu.edu>
**
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include "php.h"
#include "php_ini.h"
#include "php_safe_shell_exec.h"

#define MAX_ARGS 256
#define ERROR_BUF_SZ 512
#define RETURN_BUF_SZ 8192
#define MERGE_STDERR 1<<0
#define DISCARD_STDERR 1<<1

static function_entry safe_shell_exec_functions[] = {
    PHP_FE(safe_shell_exec, NULL)
    {NULL, NULL, NULL}
};

zend_module_entry safe_shell_exec_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    PHP_SAFE_SHELL_EXEC_EXTNAME,
    safe_shell_exec_functions,
    PHP_MINIT(safe_shell_exec),
    PHP_MSHUTDOWN(safe_shell_exec),
    NULL,
    NULL,
    NULL,
#if ZEND_MODULE_API_NO >= 20010901
    PHP_SAFE_SHELL_EXEC_VERSION,
#endif
    STANDARD_MODULE_PROPERTIES
};

PHP_INI_BEGIN()
PHP_INI_END()

PHP_MINIT_FUNCTION(safe_shell_exec)
{
  REGISTER_INI_ENTRIES();
  REGISTER_LONG_CONSTANT( "MERGE_STDERR", MERGE_STDERR, CONST_PERSISTENT | CONST_CS );
  REGISTER_LONG_CONSTANT( "DISCARD_STDERR", DISCARD_STDERR, CONST_PERSISTENT | CONST_CS );
  return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(safe_shell_exec)
{
    UNREGISTER_INI_ENTRIES();
    return SUCCESS;
}

#ifdef COMPILE_DL_SAFE_SHELL_EXEC
ZEND_GET_MODULE(safe_shell_exec)
#endif


/*++
 * Function:     PHP_FUNCTION(safe_shell_exec)
 *
 * Purpose:      implement the safe_shell_exec PHP function
 *
 * Parameters:   array -- The argument vector to execute
 *               long (optional) -- flags
 *
 * Returns:      string -- The output of the command that was executed
 *
 * Authors:      Dave McMurtrie <dave64@andrew.cmu.edu>
 *
 * Notes:        Current accepted flags are:
 *               MERGE_STDERR - The equivalent of 2>&1
 *               DISCARD_STDERR - The equivalent of 2>/dev/null 
 *--
 */
PHP_FUNCTION(safe_shell_exec)
{
  zval *argarray;
  zval **data;
  int argcount;
  HashTable *arr_hash;
  HashPosition hash_ptr;
  long flags;
  char *cmdv[MAX_ARGS];
  unsigned int cmdc;
  char error[ERROR_BUF_SZ];
  char retbuf[RETURN_BUF_SZ];
  int rc;
  int i;
  pid_t pid;
  pid_t waited;
  int status;
  int fd[2];
  int stderr_fd;
  
  memset( cmdv, '\0', MAX_ARGS );
  retbuf[0] = '\0';
  flags = 0;
 
  if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|l", &argarray, &flags ) == FAILURE) 
  {
    RETURN_STRING("Error: failed to parse parameters", 1);
  }

  arr_hash = Z_ARRVAL_P( argarray );
  argcount = zend_hash_num_elements( arr_hash );
  
  if ( argcount >= MAX_ARGS )
  {
    snprintf( error, ERROR_BUF_SZ - 1, "Error: Number of command arguments provided (%d) exceeds maximum of %d.", argcount, MAX_ARGS - 1 );
    RETURN_STRING( (char *)error, 1 );
  }
  
  /*
   * Now walk the zval array and extract the char pointers of each string data element to populate our
   * argument vector
   */
  i = 0;
  for( zend_hash_internal_pointer_reset_ex( arr_hash, &hash_ptr );
       zend_hash_get_current_data_ex( arr_hash, (void**) &data, &hash_ptr ) == SUCCESS;
       zend_hash_move_forward_ex( arr_hash, &hash_ptr ) 
     ) 
  {
    if (Z_TYPE_PP(data) == IS_STRING) 
    {
      cmdv[i++] = Z_STRVAL_PP(data);
    }
  }
  
  if ( i != argcount )
  {
    snprintf( error, ERROR_BUF_SZ - 1, "Error: Internal sanity check failed: argcount was %d, but only parsed %d string arguments", argcount, i );
    RETURN_STRING( (char *)error, 1 );
  }
   
  rc = pipe( fd );
  if ( rc == -1 )
  {
    snprintf( error, ERROR_BUF_SZ - 1, "Error: pipe() failed: %s", strerror( errno ) );
    RETURN_STRING( (char *)error, 1 );
  }
        
  if( ( pid = fork() ) == -1 )
  {
    close( fd[0] );
    close( fd[1] );
    snprintf( error, ERROR_BUF_SZ - 1, "Error: fork() failed: %s", strerror( errno ) );
    RETURN_STRING( (char *)error, 1 );
  }

  if( pid )
  {
    /*
     * in the parent.  close the write end of our pipe.
     * and read output from the child.
     */
    close( fd[1] );
    rc = read( fd[0], retbuf, sizeof retbuf - 1 );
    if ( rc == -1 )
    {
      close( fd[0] );
      snprintf( error, ERROR_BUF_SZ - 1, "Error: read from child process failed: %s", strerror( errno ) );
      RETURN_STRING( (char *)error, 1 );
    }
    close( fd[0] );

    /* explicitly NULL-terminate the return buffer */
    retbuf[rc] = '\0';

    /* wait for our child to exit */
    waited = waitpid( pid, &status, 0 );
    if ( waited == -1 )
    {
      snprintf( error, ERROR_BUF_SZ - 1, "Error: waitpid failed: %s", strerror( errno ) );
      RETURN_STRING( (char *)error, 1 );
    }

    RETURN_STRING( (char *)retbuf, 1 );
  }
  else
  {
    /*
     * in the child.  close the read end of our pipe.
     * open stdout on the write end of the pipe
     */
    close( fd[0] );
    rc = dup2( fd[1], 1 );
    if ( rc == -1 )
    {
      snprintf( error, ERROR_BUF_SZ - 1, "Error: failed to dup stdout fd: %s", strerror( errno ) );
      RETURN_STRING( (char *)error, 1 );
    }

    /*
     * Deal with stderr if any flags were passed in
     */
    if ( flags & MERGE_STDERR )
    {    
      rc = dup2( fd[1], 2 );
      if ( rc == -1 )
      {
        snprintf( error, ERROR_BUF_SZ - 1, "Error: failed to dup stderr fd: %s", strerror( errno ) );
        RETURN_STRING( (char *)error, 1 );
      }
    }
    else if ( flags & DISCARD_STDERR )
    {
      stderr_fd = open( "/dev/null", O_RDWR );
      
      if ( stderr_fd == -1 )
      {
        snprintf( error, ERROR_BUF_SZ - 1, "Error: failed to open /dev/null for discard of stderr: %s", strerror( errno ) );
        RETURN_STRING( (char *)error, 1 );
      }
      
      rc = dup2( stderr_fd, 2 );
      if ( rc == -1 )
      {
        snprintf( error, ERROR_BUF_SZ - 1, "Error: failed to dup stderr fd to /dev/null: %s", strerror( errno ) );
        RETURN_STRING( (char *)error, 1 );
      }
    }     
      
    /* close stdin because we don't use it */
    close( 0 );

    /* now do the exec */
    rc = execvp( cmdv[0], cmdv );
    if ( rc == -1 )
    {
      fprintf( stderr, "Error: exec failed for %s: %s\n", cmdv[0], strerror( errno ) );
      exit(1);
    }
  }
}
